@@ -15,72 +15,6 @@
 #include "acb_modular.h"
 #include "acb_calc.h"
 
-/* ------------------------------------------------------------------------- */
-/*  Useful helper functions                                                  */
-/* ------------------------------------------------------------------------- */
-
-/* Absolute value function on R extended to a holomorphic function in the left
-   and right half planes. */
-void
-acb_holomorphic_abs(acb_ptr res, const acb_t z, int holomorphic, slong prec)
-{
-    if (!acb_is_finite(z) || (holomorphic && arb_contains_zero(acb_realref(z))))
-    {
-        acb_indeterminate(res);
-    }
-    else
-    {
-        if (arb_is_nonnegative(acb_realref(z)))
-        {
-            acb_set_round(res, z, prec);
-        }
-        else if (arb_is_negative(acb_realref(z)))
-        {
-            acb_neg_round(res, z, prec);
-        }
-        else
-        {
-            acb_t t;
-            acb_init(t);
-            acb_neg(t, res);
-            acb_union(res, z, t, prec);
-            acb_clear(t);
-        }
-    }
-}
-
-/* Floor function on R extended to a piecewise holomorphic function in
-   vertical strips. */
-void
-acb_holomorphic_floor(acb_ptr res, const acb_t z, int holomorphic, slong prec)
-{
-    if (!acb_is_finite(z) || (holomorphic && arb_contains_int(acb_realref(z))))
-    {
-        acb_indeterminate(res);
-    }
-    else
-    {
-        arb_floor(acb_realref(res), acb_realref(z), prec);
-        arb_set_round(acb_imagref(res), acb_imagref(z), prec);
-    }
-}
-
-/* Square root function on C with detection of the branch cut. */
-void
-acb_holomorphic_sqrt(acb_ptr res, const acb_t z, int holomorphic, slong prec)
-{
-    if (!acb_is_finite(z) || (holomorphic &&
-        arb_contains_zero(acb_imagref(z)) &&
-        arb_contains_nonpositive(acb_realref(z))))
-    {
-        acb_indeterminate(res);
-    }
-    else
-    {
-        acb_sqrt(res, z, prec);
-    }
-}
-
 /* ------------------------------------------------------------------------- */
 /*  Example integrands                                                       */
 /* ------------------------------------------------------------------------- */
@@ -104,7 +38,7 @@ f_floor(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
     if (order > 1)
         flint_abort();  /* Would be needed for Taylor method. */
 
-    acb_holomorphic_floor(res, z, order != 0, prec);
+    acb_real_floor(res, z, order != 0, prec);
 
     return 0;
 }
@@ -118,12 +52,7 @@ f_circle(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
 
     acb_one(res);
     acb_submul(res, z, z, prec);
-    acb_holomorphic_sqrt(res, res, order != 0, prec);
-
-    /* Rounding could give |z| = 1 + eps near the endpoints, but we assume
-       that the interval is [-1,1] which really makes f real.  */
-    if (order == 0)
-        arb_zero(acb_imagref(res));
+    acb_real_sqrtpos(res, res, order != 0, prec);
 
     return 0;
 }
@@ -172,7 +101,7 @@ f_helfgott(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
     acb_mul(res, res, z, prec);
     acb_add_si(res, res, -6, prec);
 
-    acb_holomorphic_abs(res, res, order != 0, prec);
+    acb_real_abs(res, res, order != 0, prec);
 
     if (acb_is_finite(res))
     {
@@ -333,7 +262,7 @@ f_monster(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
     acb_init(t);
 
     acb_exp(t, z, prec);
-    acb_holomorphic_floor(res, t, order != 0, prec);
+    acb_real_floor(res, t, order != 0, prec);
 
     if (acb_is_finite(res))
     {
@@ -350,7 +279,7 @@ f_monster(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
 
 /* f(z) = sech(10(x-0.2))^2 + sech(100(x-0.4))^4 + sech(1000(x-0.6))^6 */
 int
-f_wolfram(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+f_spike(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
 {
     acb_t a, b, c;
 
@@ -540,12 +469,170 @@ f_lambertw(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
     return 0;
 }
 
+/* f(z) = max(sin(z), cos(z)) */
+int
+f_max_sin_cos(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+{
+    acb_t s, c;
+
+    if (order > 1)
+        flint_abort();  /* Would be needed for Taylor method. */
+
+    acb_init(s);
+    acb_init(c);
+
+    acb_sin_cos(s, c, z, prec);
+    acb_real_max(res, s, c, order != 0, prec);
+
+    acb_clear(s);
+    acb_clear(c);
+
+    return 0;
+}
+
+/* f(z) = erf(z/sqrt(0.0002)*0.5 +1.5)*exp(-z), example provided by Silviu-Ioan Filip */
+int
+f_erf_bent(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+{
+    acb_t t;
+
+    if (order > 1)
+        flint_abort();  /* Would be needed for Taylor method. */
+
+    acb_init(t);
+
+    acb_set_ui(t, 1250);
+    acb_sqrt(t, t, prec);
+    acb_mul(t, t, z, prec);
+    acb_set_d(res, 1.5);
+    acb_add(res, res, t, prec);
+    acb_hypgeom_erf(res, res, prec);
+
+    acb_neg(t, z);
+    acb_exp(t, t, prec);
+    acb_mul(res, res, t, prec);
+
+    acb_clear(t);
+
+    return 0;
+}
+
+/* f(z) = Ai(z) */
+int
+f_airy_ai(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+{
+    if (order > 1)
+        flint_abort();  /* Would be needed for Taylor method. */
+
+
+    acb_hypgeom_airy(res, NULL, NULL, NULL, z, prec);
+
+    return 0;
+}
+
+int
+f_horror(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+{
+    acb_t s, t;
+
+    if (order > 1)
+        flint_abort();  /* Would be needed for Taylor method. */
+
+    acb_init(s);
+    acb_init(t);
+
+    acb_real_floor(res, z, order != 0, prec);
+
+    if (acb_is_finite(res))
+    {
+        acb_sub(res, z, res, prec);
+        acb_set_d(t, 0.5);
+        acb_sub(res, res, t, prec);
+        acb_sin_cos(s, t, z, prec);
+        acb_real_max(s, s, t, order != 0, prec);
+        acb_mul(res, res, s, prec);
+    }
+
+    acb_clear(s);
+    acb_clear(t);
+
+    return 0;
+}
+
+/* f(z) = sqrt(z) */
+int
+f_sqrt(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+{
+    if (order > 1)
+        flint_abort();  /* Would be needed for Taylor method. */
+
+    acb_sqrt_analytic(res, z, order != 0, prec);
+
+    return 0;
+}
+
+/* f(z) = exp(-z^2+iz) */
+int
+f_gaussian_twist(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+{
+    if (order > 1)
+        flint_abort();  /* Would be needed for Taylor method. */
+
+    acb_mul_onei(res, z);
+    acb_submul(res, z, z, prec);
+    acb_exp(res, res, prec);
+
+    return 0;
+}
+
+/* f(z) = exp(-z) Ai(-z) */
+int
+f_exp_airy(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+{
+    acb_t t;
+
+    if (order > 1)
+        flint_abort();  /* Would be needed for Taylor method. */
+
+    acb_init(t);
+    acb_neg(t, z);
+    acb_hypgeom_airy(res, NULL, NULL, NULL, t, prec);
+    acb_exp(t, t, prec);
+    acb_mul(res, res, t, prec);
+    acb_clear(t);
+
+    return 0;
+}
+
+/* f(z) = z sin(z) / (1 + cos(z)^2) */
+int
+f_sin_cos_frac(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+{
+    acb_t s, c;
+
+    if (order > 1)
+        flint_abort();  /* Would be needed for Taylor method. */
+
+    acb_init(s);
+    acb_init(c);
+
+    acb_sin_cos(s, c, z, prec);
+    acb_mul(c, c, c, prec);
+    acb_add_ui(c, c, 1, prec);
+    acb_mul(s, s, z, prec);
+    acb_div(res, s, c, prec);
+
+    acb_clear(s);
+    acb_clear(c);
+
+    return 0;
+}
 
 /* ------------------------------------------------------------------------- */
 /*  Main test program                                                        */
 /* ------------------------------------------------------------------------- */
 
-#define NUM_INTEGRALS 24
+#define NUM_INTEGRALS 32
 
 const char * descr[NUM_INTEGRALS] =
 {
@@ -554,7 +641,7 @@ const char * descr[NUM_INTEGRALS] =
     "2 int_0^{inf} 1/(1+x^2) dx   (using domain truncation)",
     "4 int_0^1 sqrt(1-x^2) dx",
     "int_0^8 sin(x+exp(x)) dx",
-    "int_0^100 floor(x) dx",
+    "int_1^101 floor(x) dx",
     "int_0^1 |x^4+10x^3+19x^2-6x-6| exp(x) dx",
     "1/(2 pi i) int zeta(s) ds  (closed path around s = 1)",
     "int_0^1 sin(1/x) dx  (slow convergence, use -heap and/or -tol)",
@@ -573,6 +660,14 @@ const char * descr[NUM_INTEGRALS] =
     "int_C wp(x)/x^(11) dx   (contour for 10th Laurent coefficient of Weierstrass p-function)",
     "N(1000) = count zeros with 0 < t <= 1000 of zeta(s) using argument principle",
     "int_0^{1000} W_0(x) dx",
+    "int_0^pi max(sin(x), cos(x)) dx",
+    "int_{-1}^1 erf(x/sqrt(0.0002)*0.5+1.5)*exp(-x) dx",
+    "int_{-10}^10 Ai(x) dx",
+    "int_0^10 (x-floor(x)-1/2) max(sin(x),cos(x)) dx",
+    "int_{-1-i}^{-1+i} sqrt(x) dx",
+    "int_0^{inf} exp(-x^2+ix) dx   (using domain truncation)",
+    "int_0^{inf} exp(-x) Ai(-x) dx   (using domain truncation)",
+    "int_0^pi x sin(x) / (1 + cos(x)^2) dx",
 };
 
 int main(int argc, char *argv[])
@@ -835,7 +930,7 @@ int main(int argc, char *argv[])
             case 15:
                 acb_set_d(a, 0.0);
                 acb_set_d(b, 1.0);
-                acb_calc_integrate(s, f_wolfram, NULL, a, b, goal, tol, options, prec);
+                acb_calc_integrate(s, f_spike, NULL, a, b, goal, tol, options, prec);
                 break;
 
             case 16:
@@ -960,6 +1055,63 @@ int main(int argc, char *argv[])
                 acb_calc_integrate(s, f_lambertw, NULL, a, b, goal, tol, options, prec);
                 break;
 
+            case 24:
+                acb_set_d(a, 0.0);
+                acb_const_pi(b, prec);
+                acb_calc_integrate(s, f_max_sin_cos, NULL, a, b, goal, tol, options, prec);
+                break;
+
+            case 25:
+                acb_set_si(a, -1);
+                acb_set_si(b, 1);
+                acb_calc_integrate(s, f_erf_bent, NULL, a, b, goal, tol, options, prec);
+                break;
+
+            case 26:
+                acb_set_si(a, -10);
+                acb_set_si(b, 10);
+                acb_calc_integrate(s, f_airy_ai, NULL, a, b, goal, tol, options, prec);
+                break;
+
+            case 27:
+                acb_set_si(a, 0);
+                acb_set_si(b, 10);
+                acb_calc_integrate(s, f_horror, NULL, a, b, goal, tol, options, prec);
+                break;
+
+            case 28:
+                acb_set_d_d(a, -1, -1);
+                acb_set_d_d(b, -1, 1);
+                acb_calc_integrate(s, f_sqrt, NULL, a, b, goal, tol, options, prec);
+                break;
+
+            case 29:
+                acb_set_d(a, 0);
+                acb_set_d(b, ceil(sqrt(goal * 0.693147181) + 1.0));
+                acb_calc_integrate(s, f_gaussian_twist, NULL, a, b, goal, tol, options, prec);
+                acb_mul(b, b, b, prec);
+                acb_neg(b, b);
+                acb_exp(b, b, prec);
+                arb_add_error(acb_realref(s), acb_realref(b));
+                arb_add_error(acb_imagref(s), acb_realref(b));
+                break;
+
+            case 30:
+                acb_set_d(a, 0);
+                acb_set_d(b, ceil(goal * 0.693147181 + 1.0));
+                acb_calc_integrate(s, f_exp_airy, NULL, a, b, goal, tol, options, prec);
+                acb_neg(b, b);
+                acb_exp(b, b, prec);
+                acb_mul_2exp_si(b, b, 1);
+                arb_add_error(acb_realref(s), acb_realref(b));
+                break;
+
+            case 31:
+                acb_zero(a);
+                acb_const_pi(b, prec);
+                acb_calc_integrate(s, f_sin_cos_frac, NULL, a, b, goal, tol, options, prec);
+                break;
+
             default:
                 abort();
             }
